/*
 * Choeps Next Generation
 * 
 * Mark Spencer <markster@marko.net>
 *
 * Copyright(C) 1999, Adtran, Inc.
 * 
 * Distributed under the terms of the GNU General Public License (GPL) Version
 *
 * Cheops Agent Main File
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "cheops.h"
#include "logger.h"
#include "event.h"
#include "discover.h"

#ifdef DEBUG_CHEOPS_AGENT
#define DEBUG DEBUG_M
#else
#define DEBUG(a)
#endif

static struct handler {
	int id;
	cheops_event_handler cb;
	char *desc;
} agent_handlers[];


struct handler agent_handlers[] = {
	{ EVENT_DIPV4, handle_discover_request, "Discover Request Handler" },
	{ -1 },
};

void usage()
{
	fprintf(stderr, "Usage: cheops-agent [-n|-l|-a]\n"
	                "          -n: Listen using network method\n"
			"          -l: Listen using local method\n"
			"          -f: Fork to background\n"
			"          -a: Listen using all methods (default)\n");
	exit(1);
}

void register_agent_handlers()
{
	struct handler *h = agent_handlers;
	while(h->id > -1) {
		/* Try to register each of our shell items, but don't replace an existing
		   handler */
		if (event_register_handler(h->id, h->cb, 0)) 
			clog(LOG_WARNING, "Unable to register handler for '%s'\n", h->desc);
		h++;
	}
}

void do_it(int local, int ipv4)
{
	int type=0;
	if (ipv4)
		type |= AGENT_TYPE_IPV4;
	if (local)
		type |= AGENT_TYPE_LOCAL;
	if (event_create_agent(type)) {
		clog(LOG_ERROR, "unable to listen.\n");
		exit(1);
	}
	register_agent_handlers();
	cheops_main();
	return;
}

int main(int argc, char *argv[])
{
	char c;
	int local=0;
	int ipv4=0;
	int all=0;
	int dofork=0;
	pid_t pid;
	while((c = getopt(argc, argv, "alnf")) >=0) {
		switch(c) {
		case 'a':
			all++;
			break;
		case 'n':
			ipv4++;
			break;
		case 'l':
			local++;
			break;
		case 'f':
			dofork++;
			break;
		default:
			usage();
		}
	}
	/* Check for garbage */
	if (optind != argc)
		usage();
	/* Default is to listen on  all methods */
	if (!all && !local & !ipv4) 
		all++;
	if (dofork) {
		switch((pid = fork())) {
		case 0:
			/* Child */
			/* Main thread */
			do_it((all || local), (all || ipv4));
			break;
		case -1:
			/* Error */
			clog(LOG_ERROR, "fork() failed: %s\n", strerror(errno));
			exit(1);
			break;
		default:
			/* Parent */
			DEBUG(clog(LOG_NOTICE, "cheops-agent launched, pid %d\n", pid));
		}
	} else
		do_it((all || local), (all || ipv4));
	exit(0);
}

