/*
 * Choeps Next Generation
 * 
 * Brent Priddy <toopriddy@mailcity.com>
 *
 * Copyright(C) 1999, Adtran, Inc.
 * 
 * Distributed under the terms of the GNU General Public License (GPL) Version
 *
 * Network discovery code
 *
 */
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <malloc.h>
#include <stdlib.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "cheops.h"
#include "io.h"
#include "logger.h"
#include "sched.h"
#include "event.h"
#include "cheops-sh.h"
#include "discover.h"
#include "ip_utils.h"

#define PING_RETRIES 2
#define PING_INTERVAL 3000
#define PING_START_INTERVAL 10

extern unsigned short inet_checksum (void *, int);

struct ping_q {
	u32	ip;
	u32	desched;
	u32	retry;
	u32	flags;
	agent 	*a;
};

/* Buffer for creating all events */
static char ebuf[65536];
static event_hdr *eh = (event_hdr *)ebuf;
static event *ee = (event *)(ebuf + sizeof(event_hdr));

static int ping_socket = -1;
static int last_time = 0;

int handle_discover_request(event_hdr *h, event *e, agent *a);
int handle_ping_recieve(int *id, int fd, short e, void *data);
int ping_send(void *data);


int handle_discover_request(event_hdr *h, event *e, agent *a)
{
	int 	ret 	= -1;
	int	time,ms;
	u32	i,finish;
	u32	start 	= e->dipv4_e.start;
	u32	end 	= e->dipv4_e.end;
	char	*errorPtr=NULL;
	char	errorStr[] = "Can't Discover 0.0.0.0\n";
	char	errorStr2[] = "Start is larger than Finish\n";
	struct ping_q *pp=NULL;
	struct timeval tv;
	struct timezone tz;
	
//	printf("i got a discover request for %s to %s\n",ip2str(start),ip2str(end));

	if(start==0 || end==0)
	{
		errorPtr=&errorStr[0];
	}

	if(ntohl(start) > ntohl(end))
	{
		errorPtr=&errorStr2[0];
	}
	
	if(errorPtr)
	{
		eh->len = htonl(sizeof(event_hdr) + sizeof(error_r) + strlen(errorPtr));
		eh->hlen = htons(sizeof(event_hdr));
		eh->type = htons(REPLY_ERROR);
		eh->flags = 0;
		ee->error_r.error = 99;
		ee->error_r.len = strlen(errorPtr);
		strcpy( ((char *)ee + sizeof(error_r)) , errorPtr);

		if (event_send(a, eh) < 0)
			clog(LOG_WARNING, "Unable to send error reply\n");
		else
			ret = 0;
		return(ret);
	}	
	
	finish = ntohl(end);
	for(i=ntohl(start);i<=finish;i++)
	{
		/* dont send to addresses with the last octet = 0x00 or 0xFF */
		if(i & 0xFF && ((i ^ 0xFF) & 0xFF))
		{
			if((pp = malloc(sizeof(struct ping_q)))==NULL)
			{
				printf("malloc returned NULL\n");
				exit(1);
			}
			pp->ip = i; /* we put it back to network order when we send */
			pp->a = a;
			pp->retry = 0;
			pp->flags = h->flags;
			
			gettimeofday(&tv,&tz);
			ms = (int)tv.tv_sec*1000 + (int)tv.tv_usec/1000;
			if(last_time == 0)
			{
				time = 1;
			}
			else
			{
				/* send a ping every PING_START_INTERVAL ms */
				if(ms - last_time > PING_START_INTERVAL)
				{
					time = 1;
				}
				else
				{
					if(ms - last_time <= 0)
					{
						time = -ms + last_time + PING_START_INTERVAL;
					}
					else
					{
						time = ms - last_time + PING_START_INTERVAL;
					}
				}
			}
			last_time = ms + time;
			pp->desched = cheops_sched_add(time,CHEOPS_SCHED_CB(ping_send),pp);
			
//			printf("added %s\n",ip2str(htonl(i)));
		}
	}	
	return(ret);
}

int handle_ping_recieve(int *id, int fd, short e, void *data)
{
	struct ip 	*ip;
	struct icmp 	*i;
	struct ping_q	*pp;	
	struct sockaddr_in sin;
	int 	hlen;
	int	len;
	int	sinlen = sizeof(sin);
	char	buf[1024];

/* XXXXXXXXXXXXXXXX   Corrupt ping replies might crash us  XXXXXXXXXXXXXXXXXXX */

	if ((len = recvfrom(fd, buf, sizeof(buf), 0,(struct sockaddr *)&sin, &sinlen)) < 0) 
	{
    		perror("network_socket_cb: recvfrom");
	}
	
	ip = (struct ip *)buf;
	hlen = ip->ip_hl << 2;

	if(len < hlen + ICMP_MINLEN) {
		fprintf(stderr, "handle_packet: too short (%d bytes) from %s\n",
			len, inet_ntoa(sin.sin_addr));
		return(1);
	}
	i = (struct icmp *)(buf + hlen);
	if((i->icmp_type == ICMP_TIME_EXCEEDED) || (i->icmp_type == ICMP_DEST_UNREACH))
		return(1);

	if(i->icmp_type != ICMP_ECHOREPLY)
		return(1);

	bcopy((buf + hlen + 8),&pp,sizeof(void *));

	if(sin.sin_addr.s_addr != htonl(pp->ip))
	{
		return(1);
	}

	pp->retry=PING_RETRIES + 1;
		
	eh->len = htonl(sizeof(event_hdr) + sizeof(dipv4_r));
	eh->hlen = htons(sizeof(event_hdr));
	eh->type = htons(REPLY_DIPV4);
	eh->flags = pp->flags;
	ee->dipv4_r.ipaddr = htonl(pp->ip);

	if (event_send(pp->a, eh) < 0)
		clog(LOG_WARNING, "Unable to send discover reply\n");
	
	return(1);
}

int ping_send(void *data)	
{
	struct ping_q		*pp=data;
	struct protoent		*proto;
	struct sockaddr_in	to;
	struct icmp 		*i;
	char 	buf[256];
	int 	len = 8 + sizeof(void *);
	int	sent;
	int	hold;
	int	*id;
	int	ret=1;
	
	if(pp->retry == 0)
	{
		pp->desched = cheops_sched_add(PING_INTERVAL,CHEOPS_SCHED_CB(ping_send),pp);
		ret=0;	
	}
	
	if(pp->retry > PING_RETRIES )
	{
		/* remove the values from memory (they might be used even after freed, 
		 * as in the case with a ping recieve) */
		pp->ip = 0;
		pp->desched = 0;
		pp->flags = 0;
		free(pp);
		return(0);
	}
	

	if(ping_socket < 0)
	{		
		if(!(proto = getprotobyname("icmp")))
		{
			printf("socket error\n");
			exit(1);
		}
		
		if((ping_socket = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0)
		{
			printf("socket error\n");
			exit(1);
		}
		hold = 48 * 1024;
		setsockopt(ping_socket,SOL_SOCKET,SO_RCVBUF,(char *)&hold,sizeof(hold)); 

		id = cheops_io_add(ping_socket,handle_ping_recieve,CHEOPS_IO_IN,NULL);
	}
		
	pp->retry++;
		
	/* send the ICMP PING */
	memset(&to, 0, sizeof(struct sockaddr_in));
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = htonl(pp->ip);
	i = (struct icmp *)buf;
	i->icmp_type = ICMP_ECHO;
	i->icmp_code = 0;
	i->icmp_cksum = htons(0);
	i->icmp_seq = htons(pp->retry);

	bcopy(&pp, &i->icmp_id, sizeof(void *));
	bcopy(&pp,&buf[8], sizeof(void *));
	i->icmp_cksum = inet_checksum(i, len);
	
	sent = sendto(ping_socket, buf, len, 0, (struct sockaddr *)&to, sizeof(to));
	if (sent < 0 || sent != len)
	{
		if (sent < 0)
			printf("init_scan_host: sendto ");
		return(ret);	
	}				
//	printf("init_scan_host: wrote %s %d chars, ret = %d\n",inet_ntoa(to.sin_addr), len, sent);
	
	return(ret);
}
